(*----------------------------------------------------------------

  Here is a collection of tactics related to the use of UNITY99
  hol library.

  Written by Wishnu Prasetya
  Dept. of Computer Science,
  Utrecht University
  Email : wishnu@cs.uu.nl

  Last update: june 1999

----------------------------------------------------------------*)

structure UNITY99tacs = struct

local open HolKernel 
           basicHol90Lib 
      in

  infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
  infixr --> ;


(*-------------------------------------------------------------------
    PRED_TAC_v3 defs lems  tries to prove a given predicate statement
 
        defs : additional definitions which may need expansion
               before first order prover is called
        lems : additional lemmas to be supplied to the first order
               prover
-------------------------------------------------------------------*)

  fun PRED_TAC_v3 defs lems =
      let
      (*-- all definitions used to first rewrite goal to first order
           form --*)
      val all_defs = defs @ 
               [UNITY99Theory.Pvalid_def, 
                upredTheory.pSEQ_DEF, upredTheory.NOT_def, 
                upredTheory.AND_def, upredTheory.OR_def, 
                upredTheory.EQUAL_def, upredTheory.IMP_def, 
                upredTheory.RES_qAND_def, upredTheory.RES_qOR_def,
                upredTheory.TT_def, upredTheory.FF_def, 
                restr_binderTheory.RES_FORALL,
                restr_binderTheory.RES_EXISTS]
      in
      (*-- (1) expand functional equalities using extensionality --*)
      CONV_TAC (DEPTH_CONV FUN_EQ_CONV)
      (*-- (2) expand all definitions --*)
      THEN REWRITE_TAC all_defs THEN BETA_TAC THEN REWRITE_TAC all_defs 
      THEN REPEAT STRIP_TAC
      THEN TRY (bossLib.PROVE_TAC lems) 
      end ;

(*-----------------------------------------------------------------------
 
  Here is is an induction tactic. 

  LEADSTO_INDUCT_v1 lems_e lems_t lems_d do the following :

    (1) find the induction variables, and automate the lamda lifting
        and subsequent induction application
    (2) meson prover is invoked at each case (ensures lift, transitivity,
        and disjunction) to solve them. The three additional arguments 
        are lists of theorems which are passed to the meson prover to
        prove its goal.
-----------------------------------------------------------------------*)

  fun MK_ABS2_CONV x y trm = 
      (SYM o DEPTH_CONV BETA_CONV o Term) `(\^x ^y. ^trm) ^x ^y` ; 

  fun LEADSTO_INDUCT_TAC_v1 lems_e lems_t lems_d (asml,g) =
      let    
      val leadsto_term =
          let 
          val ante = (rand o rator) g
          in
          if is_conj ante then (rand o rator) ante
                                else ante
          end
      val p_term = (rand o rator) leadsto_term 
      val q_term = rand leadsto_term
      in
      (REPEAT GEN_TAC
       THEN STRIP_TAC
       THEN CONV_TAC (MK_ABS2_CONV p_term q_term)
       THEN UNDISCH_TAC leadsto_term
       THEN MATCH_MP_TAC UNITY99Theory.LEADSTO_INDUCT_thm1     
       THEN CONJ_TAC 
       THENL
       [ (*-- ensures part --*)
         REWRITE_TAC [urelTheory.SUBREL_def, 
                      restr_binderTheory.RES_FORALL,
                      restr_binderTheory.RES_EXISTS ]
         THEN BETA_TAC
         THEN REPEAT STRIP_TAC 
         THEN TRY (bossLib.PROVE_TAC lems_e),
         CONJ_TAC
         THENL
         [ (*-- transitivity part --*)
           REWRITE_TAC [urelTheory.TRANS_def, 
                        restr_binderTheory.RES_FORALL,
                        restr_binderTheory.RES_EXISTS ]
           THEN BETA_TAC
           THEN REPEAT STRIP_TAC 
           THEN TRY (bossLib.PROVE_TAC lems_t),
           (*-- disjunctivity part --*)
           REWRITE_TAC [utdcTheory.LDISJ_def, restr_binderTheory.RES_FORALL,
                        restr_binderTheory.RES_EXISTS ]
           THEN BETA_TAC 
           THEN REPEAT STRIP_TAC 
           THEN TRY (bossLib.PROVE_TAC lems_d)
         ]
       ])
       (asml,g)
       end ;

  end

end



