(*----------------------------------------------------------------------------

  Written by Wishnu Prasetya
  Dept. of Computer Science,
  Utrecht University
  Email : wishnu@cs.uu.nl

  Last update: aug 1999

  A small library of simple action language :

     [1] Conditional : Guard THEN Action
     [2] Assign1 : x := E
     [3] Assign2 : x,y := E,F
     [4] Const expression

----------------------------------------------------------------------------*)
open HolKernel basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ;

open bossLib ;

val _ = new_theory "usimple_action" ;

val Then_def = Define
   `Then g a = (\(s:'s) t. (g s ==> a s t) /\ (~g s ==>(t=s)))` ;

val _ = set_fixity "Then" (Infix 200) ;   

val Then_adef1 = store_thm
  ("Then_adef1",
 --`g Then a = (\(s:'s) t. (g s /\ a s t) \/ (~g s /\ (t=s)))`--,

    REWRITE_TAC [Then_def]
    THEN CONV_TAC (DEPTH_CONV FUN_EQ_CONV) THEN BETA_TAC
    THEN CONV_TAC (DEPTH_CONV FUN_EQ_CONV) THEN BETA_TAC
    THEN PROVE_TAC[]) ;
    
val Assign1_def = Define 
   `:= var exp = 
    (\(s:'var->'val) t. (t var = exp s) /\
                        (!v. ~(v=var) ==> (t v = s v)))` ;

val _ = set_MLname ":=_def" "Assign1_def";

val _ = set_fixity ":=" (Infix 250) ;

val Assign2_def = Define 
   `::= (var1,var2) (exp1,exp2) = 
    (\(s:'var->'val) t. (t var1 = exp1 s) /\
                        (t var2 = exp2 s) /\
                        (!v. ~(v=var1) /\ ~(v=var2) 
                               ==> (t v = s v)))` ;
val _ = set_MLname "::=_def" "Assign2_def";

val _ = set_fixity "::=" (Infix 250) ;


val Assign3_def = Define 
   `:::= (var1,var2,var3) (exp1,exp2,exp3) = 
    (\(s:'var->'val) t. (t var1 = exp1 s) /\
                        (t var2 = exp2 s) /\
                        (t var3 = exp3 s) /\
                        (!v. ~(v=var1) /\ ~(v=var2) /\ ~(v=var3) 
                               ==> (t v = s v)))` ;

val _ = set_MLname ":::=_def" "Assign3_def";

val _ = set_fixity ":::=" (Infix 250) ;

val Const_def = Define `Const c = (\z:'a. c)` ;

val Assign1_CASE_SPLIT_thm1 = store_thm
  ("Assign1_CASE_SPLIT_thm1",
 --`!y. (G Then x := (E:('var->'val)->'val)) s t /\
         G s 
         ==>
        (~(y=x) /\ (t y = s y)) \/
        ( (y=x) /\ (t y = E s))`--,

   REWRITE_TAC [Then_def, Assign1_def]
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val Assign1_CASE_SPLIT_thm2 = store_thm
  ("Assign1_CASE_SPLIT_thm2",
 --`!y. (G Then x := (E:('var->'val)->'val)) s t /\
         ~G s 
         ==>
        (t y = s y)`--,

   REWRITE_TAC [Then_def, Assign1_def]
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val Assign2_CASE_SPLIT_thm1 = store_thm
  ("Assign2_CASE_SPLIT_thm1",
 --`!y. (G Then (x1,x2) ::= ((E1:('var->'val)->'val),E2)) s t /\
         G s
         ==>
        (~(y=x1) /\ ~(y=x2) /\ (t y = s y)) \/
        ((y=x1) /\ (t y = E1 s)) \/
        ((y=x2) /\ (t y = E2 s))`--,
    
   REWRITE_TAC [Then_def, Assign2_def]
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val Assign2_CASE_SPLIT_thm2 = store_thm
  ("Assign2_CASE_SPLIT_thm2",
 --`!y. (G Then (x1,x2) ::= ((E1:('var->'val)->'val),E2)) s t /\
        ~G s
         ==>
        (t y = s y)`--,
    
   REWRITE_TAC [Then_def, Assign2_def]
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val Assign3_CASE_SPLIT_thm1 = store_thm
  ("Assign3_CASE_SPLIT_thm1",
 --`!y. (G Then (x1,x2,x3) :::= ((E1:('var->'val)->'val),E2,E3)) s t /\
         G s
         ==>
        (~(y=x1) /\ ~(y=x2) /\ ~(y=x3) /\ (t y = s y)) \/
        ((y=x1) /\ (t y = E1 s)) \/
        ((y=x2) /\ (t y = E2 s)) \/
        ((y=x3) /\ (t y = E3 s))`--,
    
   REWRITE_TAC [Then_def, Assign3_def]
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val Assign3_CASE_SPLIT_thm2 = store_thm
  ("Assign3_CASE_SPLIT_thm2",
 --`!y. (G Then (x1,x2,x3) :::= ((E1:('var->'val)->'val),E2,E3)) s t /\
        ~G s
         ==>
        (t y = s y)`--,
    
   REWRITE_TAC [Then_def, Assign3_def]
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val _ = export_theory() ;