(*----------------------------------------------------------------------------

  Written by Wishnu Prasetya
  Dept. of Computer Science,
  Utrecht University
  Email : wishnu@cs.uu.nl

  Last update: aug 1999

  A list of ML functions to use with usimple_actionTheory.

----------------------------------------------------------------------------*)

structure usimple_actiontacs = struct

local open HolKernel 
           basicHol90Lib
           bossLib 
           btacs
           arithLib
           upredTheory
           UNITY99Theory
           usimple_actionTheory
      in

infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ;

fun NUM_EQ_CONV trm =
    if is_eq trm 
       then 
         let
         val left_term  = #lhs (dest_eq trm)
         val right_term = #rhs (dest_eq trm)  
         in
         if ((type_of left_term = Type `:num`) andalso
             (is_const left_term) andalso
             (is_const right_term))
             then (NEGATE_CONV ARITH_CONV) trm
             else NO_CONV trm
         end
      else NO_CONV trm ;


(*-- extract the guard term, assignment term, etc 
     assume aterm is of the form G Then A --*)

fun is_Assign1 aterm =
    ((#Name o dest_const o rator o rator o rand) aterm
     =
     ":=") ;

fun is_Assign2 aterm =
    ((#Name o dest_const o rator o rator o rand) aterm
     =
     "::=") ;

fun getGuard aterm = (rand o rator) aterm ;

fun getVarA1 aterm  = (rand o rator o rand) aterm ;
fun getVarA2 aterm  = (dest_pair o rand o rator o rand) aterm ;
fun getExprA1 aterm = (rand o rand) aterm ;
fun getExprA2 aterm = (dest_pair o rand o rand) aterm ;

(*--      assume goal of the form [..., P s, A s t] ==> Q s --*)

fun GUARD_SPLIT_TAC_v1 (asml,c) =
    let
    val guard_term = 
        mk_comb {Rator = (getGuard o rator o rator o el 1) asml,
                 Rand  = (rand o rator o el 1) asml}
    in
    ASM_CASES_TAC guard_term (asml,c) 
    end ; 


val LTHEN = foldr (fn (x,r) => x THEN r) ALL_TAC ;

fun Assign_G_TAC_v1 v_list v_defs (asml,g) =
    let
    val A_term = (ASSUME o el 2) asml
    val G_term = ASSUME (el 1 asml)
    val SPLIT_thm = 
           if (is_Assign1 o rator o rator o concl) A_term
              then Assign1_CASE_SPLIT_thm1
              else if (is_Assign2 o rator o rator o concl) A_term
                      then Assign2_CASE_SPLIT_thm1
                      else Assign3_CASE_SPLIT_thm1

    fun new_term v = (REWRITE_RULE [] o
                   CONV_RULE (DEPTH_CONV NUM_EQ_CONV) o
                   REWRITE_RULE v_defs o
                   ISPEC v o
                   MATCH_MP SPLIT_thm)
                   (CONJ A_term G_term)
      
    in
      (LTHEN (map (ASSUME_TAC o new_term) v_list)
       THEN DEL_TAC (concl A_term))
      (asml,g)
    end ;


fun Assign_notG_TAC_v1 v_list v_defs (asml,g) =
    let
    val A_term = (ASSUME o el 2) asml
    val G_term = ASSUME (el 1 asml)
    val SPLIT_thm = 
           if (is_Assign1 o rator o rator o concl) A_term
              then Assign1_CASE_SPLIT_thm2
              else if (is_Assign2 o rator o rator o concl) A_term
                      then Assign2_CASE_SPLIT_thm2
                      else Assign3_CASE_SPLIT_thm2


    fun new_term v = (ISPEC v o
                      MATCH_MP SPLIT_thm)
                      (CONJ A_term G_term)
      
    in
      (LTHEN (map (ASSUME_TAC o new_term) v_list)
       THEN DEL_TAC (concl A_term))
       (asml,g)
    end ;


fun Assign_BREAK_TAC_v1 v_list v_defs =
    GUARD_SPLIT_TAC_v1 
    THENL
    [Assign_G_TAC_v1 v_list v_defs,
     Assign_notG_TAC_v1 v_list v_defs ] ;


fun Assign_BREAK_TAC_v1 v_list v_defs =
    GUARD_SPLIT_TAC_v1 
    THENL
    [Assign_G_TAC_v1 v_list v_defs,
     Assign_notG_TAC_v1 v_list v_defs ] ;


val pred_operators_defs =
   [TT_def, FF_def, NOT_def, AND_def, OR_def, IMP_def,
    EQUAL_def, RES_qAND_def, RES_qOR_def] ;    


val ZERO_pOR = prove
  (
 --`(p:'a->bool) OR TT = TT`--,
    REWRITE_TAC pred_operators_defs
    THEN BETA_TAC THEN PROVE_TAC[]) ;

val ZERO_pAND = prove
  (
 --`(p:'a->bool) AND FF = FF`--,
    REWRITE_TAC pred_operators_defs
    THEN BETA_TAC THEN PROVE_TAC[]) ;

val UNIT_pOR = prove
  (
 --`(p:'a->bool) OR FF = p`--,
    REWRITE_TAC pred_operators_defs
    THEN BETA_TAC THEN PROVE_TAC[]) ;

val UNIT_pAND = prove
  (
 --`(p:'a->bool) AND TT = p`--,
    REWRITE_TAC pred_operators_defs
    THEN BETA_TAC THEN PROVE_TAC[]) ;

val NOT_NOT = prove
  (
 --`NOT (NOT (p:'a->bool)) = p`--,
    REWRITE_TAC pred_operators_defs
    THEN BETA_TAC THEN PROVE_TAC[]) ;

val EQUAL_TO_IMP = prove
  (
 --`(p:'a->bool) EQUAL q = (p IMP q) AND (q IMP p)`--,
    REWRITE_TAC pred_operators_defs
    THEN BETA_TAC THEN PROVE_TAC[]) ;

val IMP_TO_OR = prove
  (
 --`(p:'a->bool) IMP q = (NOT p) OR q`--,
    REWRITE_TAC pred_operators_defs
    THEN BETA_TAC THEN PROVE_TAC[]) ;

(*-- a normalizer for predicates : removing all EQUAL and IMP,
     and simplifying NOTS.
     note: normalizer for quantified predicates is not
           implemented yet! --*)

val Pred_NORMALIZER_TAC_v1 =
    REWRITE_TAC [(*-- eliminates EQUAL and IMP --*)
                 EQUAL_TO_IMP, IMP_TO_OR, 
                 (*-- simplify NOTs --*)
                 NOT_NOT, pDEMORGAN1, pDEMORGAN2,
                 pOR_ZERO, ZERO_pOR, pAND_ZERO, ZERO_pAND,
                 pOR_UNIT, UNIT_pOR, pAND_UNIT, UNIT_pAND ] ;
                 

val HOA_BREAK_TAC_v1 =
    (*-- normalizing/simplyfing hoare triples based on the form of
         the pre and post predicates.
         NOTE: Normalizer for quantified predicates is not implemented
         yet. --*)
  
    Pred_NORMALIZER_TAC_v1          
    THEN REWRITE_TAC [HOA_OR, HOA_AND]
    THEN REPEAT CONJ_TAC ;


fun UNLESS_BREAK_TAC_v1 
    program_def
    actions_defs
    vars_list
    vars_defs
    vals_defs
    =
    REWRITE_TAC ([STABLE_def, UNLESS_def, 
                  restr_binderTheory.RES_FORALL] @
                 [program_def] @ actions_defs)
   THEN REPEAT STRIP_TAC
   THEN POP_ASSUM SUBST1_TAC
   THEN (CONV_TAC BETA_CONV)
   THEN HOA_BREAK_TAC_v1
   THEN REWRITE_TAC [HOA_def]
   THEN REPEAT STRIP_TAC 
   THEN TRY (Assign_BREAK_TAC_v1 vars_list vars_defs
             THEN UNDISCH_ALL_TAC
             THEN REWRITE_TAC ([Const_def] @ vars_defs @ vals_defs @
                               pred_operators_defs)
             THEN BETA_TAC
             THEN RW_TAC bool_ss [])  ;


end
end