open HolKernel basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ;

val _ = new_theory "uvars" ;
 
(*
load "bossLib" ;
load "upredTheory" ;
load "restr_binderTheory" ;
*)

open bossLib ;
open restr_binderTheory ;
open upredTheory ;
open btacs ;


val State  = ty_antiq (Type `:'var -> 'val`) ;
val Pred   = ty_antiq (Type `:^State -> bool`) ;
val Action = ty_antiq (Type `:^State -> ^State -> bool`) ;

val _ = associate_restriction ("!!", "RES_qAND") ;
val _ = associate_restriction ("??", "RES_qOR") ;

fun XRULE_ASSUM_TAC rule = RULE_ASSUM_TAC (liteLib.orelsef (rule,I)) ;

(*--------------------------------------------------------------------------
  val Nov (No Value) denote a polymorphic, fixed, but for the rest Not 
  important value
--------------------------------------------------------------------------*)

val _ = new_constant {Name = "Nov", Ty= (Type `:'a`)} ;

val Pj_def = Define `Pj (f:'a->'b) A x = ((A x) => f x | Nov)` ;
val _ = set_fixity "Pj" (Infix 500) ;

val Pj_EQ = store_thm
  ("Pj_EQ",
   Term `!(f:'a->'b) g A. ((f Pj A) = (g Pj A)) = !x. A x ==> (f x = g x)`,
   (CONV_TAC  o ONCE_DEPTH_CONV) FUN_EQ_CONV
   THEN REWRITE_TAC [Pj_def]
   THEN REPEAT GEN_TAC 
   THEN EQ_TAC THEN REPEAT STRIP_TAC
   THENL 
   [ (*- (1) -*)
     XRULE_ASSUM_TAC (SPEC (Term `x:'a`))
     THEN XRULE_ASSUM_TAC
         (fn thm=> (CONV_RULE o ONCE_DEPTH_CONV) COND_CONV 
                (REWRITE_RULE [ASSUME (Term `(A:'a->bool) x`)] thm)) 
     THEN ONCE_ASM_REWRITE_TAC[] 
     THEN REFL_TAC ,
    (*- (2) -*)
     ASM_CASES_TAC (Term `(A:'a->bool) x`)
     THEN ASM_REWRITE_TAC []
     THEN RES_TAC ]) ;

val Pj_Pj = store_thm
  ("Pj_Pj",
   --`((f:'a->'b) Pj A) Pj A = f Pj A`--,
   REWRITE_TAC [Pj_EQ]
   THEN REWRITE_TAC [Pj_def]
   THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]) ;

val Pj_COMPO = store_thm
  ("Pj_COMPO",
 --`((f:'a->'b) Pj A) Pj B = f Pj (B AND A)`--,
    EXT_TAC
    THEN REWRITE_TAC [Pj_def, AND_def]
    THEN BETA_TAC
    THEN COND_CASES_TAC
    THEN COND_CASES_TAC
    THEN ASM_REWRITE_TAC[]
    THEN PROVE_TAC[]) ;

(*--------------------------------------------------------------------------
  A set A is said to confine a function f:('a->'b)->'c if f is ignorant to
  all part of NOT A of any s:'a->'b.
  
  So, a variable set A confines a predicate p if p does not restrict the
  value of any variable v outside A.
--------------------------------------------------------------------------*)

val CONF_def  = Define 
   `CONF A (p:('a->'b)->'c) = !s t. (s Pj A = t Pj A) ==> (p s = p t)` ;

val _ = set_fixity "CONF" (Infix 480) ;

(*----------------------------------------------------------------------
  lifting projection to state predicate level 
------------------------------------------------------------------------*)

val pj_def = Define 
    `pj (p:^Pred) V = (\s. p (s Pj V))` ;

val _ = set_fixity "pj" (Infix 500) ;

val pj_pj = store_thm
  ("pj_pj",
 --`((p:^Pred) pj V) pj V = p pj V`--,
    CONV_TAC FUN_EQ_CONV
    THEN REWRITE_TAC [pj_def]
    THEN BETA_TAC THEN REWRITE_TAC [Pj_Pj] ) ;

val pj_COMPO = store_thm
  ("pj_COMPO",
 --`((p:^Pred) pj V) pj W = p pj (V AND W)`--,
    EXT_TAC
    THEN REWRITE_TAC [pj_def]
    THEN BETA_TAC
    THEN REWRITE_TAC [Pj_COMPO]) ;

(*----------------------------------------------------------------------
   an alternate definition for confinement
----------------------------------------------------------------------*)

val CONF_adef1 = store_thm
  ("CONF_adef1",
   --`(V CONF (p:^Pred)) = (p = (p pj V))`--,
   REWRITE_TAC [CONF_def, pj_def]
   THEN (CONV_TAC  o ONCE_DEPTH_CONV) FUN_EQ_CONV
   THEN BETA_TAC
   THEN EQ_TAC THEN REPEAT STRIP_TAC
   THENL
     [ FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC [Pj_Pj] ,
       SUBGOAL_THEN (--`(s:^State) Pj V = t Pj V`--)
                    ASSUME_TAC
       THENL
         [ CONV_TAC FUN_EQ_CONV THEN FIRST_ASSUM ACCEPT_TAC ,
           ALL_TAC ] 
       THEN PROVE_TAC[] 
      ]
  ) ;

(*----------------------------------------------------------------------
   projection distributes over predicate operators.
----------------------------------------------------------------------*)
       
val pj_AND = store_thm
  ("pj_AND",
   --`(p AND q) pj V = ((p:^Pred) pj V) AND (q pj V)`--,
   CONV_TAC FUN_EQ_CONV
   THEN REWRITE_TAC [AND_def, pj_def] 
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val pj_OR = store_thm
  ("pj_OR",
   --`(p OR q) pj V = ((p:^Pred) pj V) OR (q pj V)`--,
   CONV_TAC FUN_EQ_CONV
   THEN REWRITE_TAC [OR_def, pj_def] 
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val pj_IMP = store_thm
  ("pj_IMP",
   --`(p IMP q) pj V = ((p:^Pred) pj V) IMP (q pj V)`--,
   CONV_TAC FUN_EQ_CONV
   THEN REWRITE_TAC [IMP_def, pj_def] 
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val pj_EQUAL = store_thm
  ("pj_EQUAL",
   --`(p EQUAL q) pj V = ((p:^Pred) pj V) EQUAL (q pj V)`--,
   CONV_TAC FUN_EQ_CONV
   THEN REWRITE_TAC [EQUAL_def, pj_def] 
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val pj_NOT = store_thm
  ("pj_NOT",
   --`(NOT p) pj V = NOT ((p:^Pred) pj V)`--,
   CONV_TAC FUN_EQ_CONV
   THEN REWRITE_TAC [NOT_def, pj_def] 
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val pj_qOR = store_thm
  ("pj_qOR",
 --`(??i::W. (i:^Pred)) pj V  = (??i::W. i pj V)`--,
   CONV_TAC FUN_EQ_CONV
   THEN REWRITE_TAC [RES_qOR_def, pj_def]
   THEN RESTRICT_DEF_TAC THEN PROVE_TAC[]) ;  

val pj_qOR_GEN = store_thm
  ("pj_qOR_GEN",
 --`(??i::W. (f:'index->^Pred) i) pj V  = (??i::W. (f i) pj V)`--,
   CONV_TAC FUN_EQ_CONV
   THEN REWRITE_TAC [RES_qOR_def, pj_def]
   THEN RESTRICT_DEF_TAC THEN PROVE_TAC[]) ; 

val pj_qAND = store_thm
  ("pj_qAND",
 --`(!!i::W. (i:^Pred)) pj V  = (!!i::W. i pj V)`--,
   CONV_TAC FUN_EQ_CONV
   THEN REWRITE_TAC [RES_qAND_def, pj_def]
   THEN RESTRICT_DEF_TAC THEN PROVE_TAC[]) ;    

val pj_qAND_GEN = store_thm
  ("pj_qAND_GEN",
 --`(!!i::W. (f:'index->^Pred) i) pj V  = (!!i::W. (f i) pj V)`--,
   CONV_TAC FUN_EQ_CONV
   THEN REWRITE_TAC [RES_qAND_def, pj_def]
   THEN RESTRICT_DEF_TAC THEN PROVE_TAC[]) ; 

(*--------------------------------------------------------------------------
  A variable set V is ignored by an action A iff A does not change any
  variable v in V.
--------------------------------------------------------------------------*)

val IG_BY_def = Define
   `IG_BY V (A:^Action) = !s t. A s t ==> (s Pj V = t Pj V)` ;

val _ = set_fixity "IG_BY" (Infix 460) ;

(*--------------------------------------------------------------------------
  A variable set V is invicilble to an action A iff A does not depend on
  the value of any variable in V.
--------------------------------------------------------------------------*)

val INVI_def =  Define
   `INVI V (A:^Action) =
    !s t s' t'. (s Pj (NOT V) = s' Pj (NOT V)) /\
                (t Pj (NOT V) = t' Pj (NOT V)) /\
                (s' Pj V = t' Pj V) /\
                A s t
                ==> A s' t'` ;

val _ = set_fixity "INVI" (Infix 460) ;


(*--------------------------------------------------------------------------
          monotonicity properties of Pj, CONF, IG_BY, and INVI
--------------------------------------------------------------------------*)

val Pj_EQ_MONO = store_thm 
  ("Pj_EQ_MONO",
   Term `!A B (f:'a->'b) g.
         |== (A IMP B) /\ ((f Pj B) = (g Pj B)) ==> ((f Pj A) = (g Pj A))`,
   REWRITE_TAC [IMP_def, pSEQ_DEF, Pj_EQ]
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val CONF_MONO = store_thm 
  ("CONF_MONO",
   Term `!A B (p:('a->'b)->'c). |== (A IMP B) /\ (A CONF p) ==> (B CONF p)`,
    REWRITE_TAC [CONF_def, IMP_def, pSEQ_DEF, Pj_EQ]
    THEN BETA_TAC THEN PROVE_TAC[]) ;

val IG_BY_MONO = store_thm 
  ("IG_BY_MONO",
   Term `!(A:^Action) V W. |== (V IMP W) /\ (W IG_BY A) ==> (V IG_BY A)`,
   REWRITE_TAC [IG_BY_def, IMP_def, pSEQ_DEF, Pj_EQ]
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val INVI_MONO = store_thm 
  ("INVI_MONO",
   Term `!(A:^Action) V W. 
        |== (V IMP W) /\ (W IG_BY A) /\ (W INVI A) ==> (V INVI A)`,
    REWRITE_TAC [INVI_def, IG_BY_def, NOT_def, IMP_def, pSEQ_DEF, Pj_EQ]
    THEN BETA_TAC 
    THEN REPEAT STRIP_TAC
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN EXISTS_TAC (Term `s:^State`)
    THEN EXISTS_TAC (Term `t:^State`)
    THEN PROVE_TAC[]) ;

(*--------------------------------------------------------------------------
            confinement is preserved by predicate operators
--------------------------------------------------------------------------*)

val CONF_TT = store_thm 
  ("CONF_TT",
   Term `!V. V CONF (TT:^Pred)`,
   REWRITE_TAC [CONF_def, TT_def] 
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val CONF_FF = store_thm 
  ("CONF_FF",
   Term `!V. V CONF (FF:^Pred)`,
   REWRITE_TAC [CONF_def, FF_def, Pj_EQ] 
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val CONF_NOT = store_thm 
  ("CONF_NOT",
   Term `!V (p:^Pred). V CONF p ==> V CONF (NOT p)`,
   REWRITE_TAC [CONF_def, NOT_def, Pj_EQ] 
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val CONF_AND = store_thm 
  ("CONF_AND",
   Term `!V p (q:^Pred). (V CONF p) /\ (V CONF q) ==> V CONF (p AND q)`,
   REWRITE_TAC [CONF_def, AND_def, Pj_EQ]
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val CONF_OR = store_thm ("CONF_OR",
   Term `!V p (q:^Pred). (V CONF p) /\ (V CONF q) ==> V CONF (p OR q)`,
   REWRITE_TAC [CONF_def, OR_def, Pj_EQ]
   THEN BETA_TAC THEN PROVE_TAC[]) ;

val CONF_IMP = store_thm 
  ("CONF_IMP",
   Term `!V p (q:^Pred). (V CONF p) /\ (V CONF q) ==> V CONF (p IMP q)`,
   REWRITE_TAC [CONF_def, IMP_def, Pj_EQ]
   THEN BETA_TAC THEN PROVE_TAC[]) ;


val CONF_qAND = store_thm 
  ("CONF_qAND",
   Term `!V W (P:'ind->^Pred).
        (!i:'ind::W. V CONF (P i)) ==> V CONF (!!i::W.(P i))`,
     REWRITE_TAC [CONF_def, RES_qAND_def, RES_FORALL, Pj_EQ]
     THEN BETA_TAC THEN PROVE_TAC[]) ;

val CONF_qOR = store_thm 
  ("CONF_qOR",
   Term `!V W (P:'ind->^Pred).
       (?i. W i) /\ (!i::W. V CONF (P i)) ==> V CONF (??i::W. P i)`,
   REWRITE_TAC [CONF_def, RES_qOR_def, RES_FORALL, RES_EXISTS, Pj_EQ]
   THEN BETA_TAC THEN PROVE_TAC[]) ;  

val _ = export_theory() ;



